"""xbrl_parser - Parser for parsing XBRL and iXBRL files (instance documents, taxonomy schemas, taxonomy linkbases)."""

"""
This package contains all classes and methods for parsing both XBRL and iXBRL
The Basics of XBRL are: Concepts, Taxonomies, Values, Contexts, Facts, Instances and Dimension's

Concepts:
A Concept is the definition of a defined term in a particular domain. For example the concept profit should be displayed
as “loss” when there were more expenses than income.

Taxonomies:
A Taxonomy is a collection of related concepts

Value:
A value is a numeric, boolean or text information.

Context:
A context defines a reference point.

Fact:
A Fact is a combination of a concept from a taxonomy, with a value and a context.
For example value 1.5 with the concept EarningsPerShare from the taxonomy us-gaap/2018 and the context From 2018-12-31
to 2019-03-31 is a Fact.

Instance:
A instance is a collection of a facts. New instance documents are created for every new period.

Dimensions:
Some companies use dimensions to differentiate facts with the same context. For example if you want to show revenue over
the same time frame but for different continents.
"""


class XbrlParseException(Exception):
    """
    Generic Class representing a exception occurred while parsing xbrl
    """
    pass


class LinkbaseNotFoundException(XbrlParseException):
    """ Generic exception for imported linkbases that could not be fetched locally or remote """
    pass


class TaxonomyParseException(XbrlParseException):
    """
    Generic class for a exception thrown while parsing a taxonomy
    """
    pass


class TaxonomyNotFound(TaxonomyParseException):
    """
    Raised when the corresponding taxonomy from a namespace was not found.
    i.e the filer used the tag dei:DocumentEntityInformation but did not import the dei taxonomy in the schema file
    """

    def __init__(self, namespace) -> None:
        super().__init__(
            "The taxonomy with namespace {} could not be found. Please check if it is imported in the schema file".format(
                namespace))


class InstanceParseException(XbrlParseException):
    """
    Generic class for an exception thrown while parsing an xbrl instance file
    """
    pass


class ContextParseException(InstanceParseException):
    """
    Exception thrown when a Context could not be parsed
    """
    pass


__version__ = '2.2.0'
__author__ = 'Manuel Schmidt <hello@schmidt-manuel.de>'
__all__ = [
    XbrlParseException,
    TaxonomyParseException,
    TaxonomyNotFound,
    InstanceParseException,
    ContextParseException,
]



"""
Downloads files and stores them locally.
"""
import os
import re
import zipfile
from pathlib import Path

from xbrl.helper.connection_manager import ConnectionManager


class HttpCache:
    """
    Simple persistent HTTP cache. Requests files over http and stores them into the cache. Just returns
    the file path if the same file is requested twice. Also automatically handles retries when request fails.
    """

    def __init__(self, cache_dir: str, delay: int = 500, verify_https: bool = True):
        """
        :param cache_dir: Root directory of the disk cache (all requested files will be cached in this directory)
        :param delay: Minimum time in milliseconds between two requests
        :param verify_https: Disable SSL certificate validation for speed up (see https://github.com/manusimidt/py-xbrl/pull/57)
        """
        # check if the cache_dir ends with a /
        if not cache_dir.endswith('/'): cache_dir += '/'
        self.cache_dir: str = cache_dir
        self.headers: dict or None = None
        self.connection_manager = ConnectionManager(delay, verify_https=verify_https)

    def set_headers(self, headers: dict) -> None:
        """
        Sets the header for all following request

        :param headers: python dictionary with string key and value
        Example header:

        .. code-block:: json

            {
                "From": "pete.smith@example.com",
                "User-Agent" : "ExampleBot/1.0 (https.example.com/exampleBot)"
            }
        """
        self.headers = headers
        self.connection_manager._headers = headers

    def set_connection_params(self, delay: int = 500, retries: int = 5, backoff_factor: float = 0.8,
                              logs: bool = True) -> None:
        """
        Sets the connection params for all following request

        :param delay: Minimum time in milliseconds between two requests
        :param retries: int specifying how many times a request will be tried before assuming its failure.
        :param backoff_factor: Used to measure time to sleep between failed requests. The formula used is:
            {backoff factor} * (2 ** ({number of total retries} - 1))
        :param logs: enables or disables download logs
        """
        self.connection_manager._delay_ms = delay
        self.connection_manager._retries = retries
        self.connection_manager._backoff_factor = backoff_factor
        self.connection_manager.logs = logs

    def cache_file(self, file_url: str) -> str:
        """
        Caches a file in the http cache.

        :param file_url: url (https link) to the file to be cached.
        :return: returns the absolute path to the cached file
        """
        file_path: str = self.url_to_path(file_url)
        # first check if the files
        if os.path.exists(file_path):
            return file_path

        file_dir_path: str = '/'.join(file_path.split('/')[0:-1])
        # try to download the file
        if not os.path.isdir(file_dir_path):
            os.makedirs(file_dir_path)

        query_response = self.connection_manager.download(file_url, headers=self.headers)

        if not query_response.status_code == 200:
            if query_response.status_code == 404:
                raise Exception(
                    "Could not find file on {}. Error code: {}".format(file_url, query_response.status_code))
            else:
                raise Exception(
                    "Could not download file from {}. Error code: {}".format(file_url, query_response.status_code))

        with open(file_path, "wb+") as file:
            file.write(query_response.content)
            file.close()

        return file_path

    def purge_file(self, file_url: str) -> bool:
        """
        Removes a file from the cache

        :param file_url: url (https link) to the file to be deleted.
        :return: true if the file was deleted, false if it could not be found
        """
        try:
            os.remove(self.url_to_path(file_url))
        except FileNotFoundError:
            return False
        return True

    def url_to_path(self, url: str) -> str:
        """
        Takes a url and converts it to the absolute local cache path

        :param url: url of the file you want to know the cache path
        :return: absolute local cache path
        """
        return self.cache_dir + re.sub("https?://", "", url)

    def cache_edgar_enclosure(self, enclosure_url: str) -> str:
        """
        Downloads the ZIP folder, extracts it and stores the files in the cache.

        :param enclosure_url: url to the zip folder.
        :return: relative path to extracted zip's content
        """
        # todo: why is it called "cache_edgar_enclosure" you could theoretically cache any zip enclosure.
        if not enclosure_url.endswith('.zip'):
            raise Exception("This is not a valid zip folder")
        # download the zip folder and store it into the default http cache
        enclosure_path = self.cache_file(file_url=enclosure_url)
        submission_dir_path = self.url_to_path('/'.join(enclosure_url.split('/')[:-1]))
        # extract the zip folder
        with zipfile.ZipFile(enclosure_path, "r") as zip_ref:
            zip_ref.extractall(submission_dir_path)
            zip_ref.close()
        return submission_dir_path

    @DeprecationWarning
    def find_entry_file(self, dir_path: str) -> str or None:
        """
        NOTE: This function only works for enclosed SEC submissions that where already downloaded!
        Also this function does only return the most likely file path for the instance document.
        If you want to be certain i would recommend to use the SEC Structured Disclosure RSS Feeds
        https://www.sec.gov/structureddata/rss-feeds-submitted-filings
        These rss feeds list all files per submission and gives you information about the filetype (instance document,
        taxonomy schema, label linkbase, exhibit e.t.c)

        Find the most likely entry file in provided filling directory
        """

        # filter for files in interest
        valid_files = []
        for ext in '.htm .xml .xsd'.split():  # valid extensions in priority
            for f in os.listdir(dir_path):
                f_full = os.path.join(dir_path, f)
                if os.path.isfile(f_full) and f.lower().endswith(ext):
                    valid_files.append(f_full)

        # find first file which is not included by others
        entry_candidates = []
        for file1 in valid_files:
            f_dir, file_nm = os.path.split(file1)
            # foreach file check all other for inclusion
            found_in_other = False
            for file2 in valid_files:
                if file1 != file2:
                    if file_nm in Path(file2).read_text():
                        found_in_other = True
                        break

            if not found_in_other:
                entry_candidates.append((file1, os.path.getsize(file1)))

        # if multiple choose biggest
        entry_candidates.sort(key=lambda tup: tup[1], reverse=True)
        if len(entry_candidates) > 0:
            file_path, size = entry_candidates[0]
            return file_path
        return None



# This library is a simple implementation of a function to convert textual
# numbers written in English into their integer representations.
#
# This code is open source according to the MIT License as follows.
#
# Copyright (c) 2008 Greg Hewgill
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# https://github.com/ghewgill/text2num/

import re

Small = {
    'zero': 0,
    'one': 1,
    'two': 2,
    'three': 3,
    'four': 4,
    'five': 5,
    'six': 6,
    'seven': 7,
    'eight': 8,
    'nine': 9,
    'ten': 10,
    'eleven': 11,
    'twelve': 12,
    'thirteen': 13,
    'fourteen': 14,
    'fifteen': 15,
    'sixteen': 16,
    'seventeen': 17,
    'eighteen': 18,
    'nineteen': 19,
    'twenty': 20,
    'thirty': 30,
    'forty': 40,
    'fifty': 50,
    'sixty': 60,
    'seventy': 70,
    'eighty': 80,
    'ninety': 90
}

Magnitude = {
    'thousand': 1000,
    'million': 1000000,
    'billion': 1000000000,
    'trillion': 1000000000000,
    'quadrillion': 1000000000000000,
    'quintillion': 1000000000000000000,
    'sextillion': 1000000000000000000000,
    'septillion': 1000000000000000000000000,
    'octillion': 1000000000000000000000000000,
    'nonillion': 1000000000000000000000000000000,
    'decillion': 1000000000000000000000000000000000,
}


class NumberException(Exception):
    def __init__(self, msg):
        Exception.__init__(self, msg)


def text2num(s) -> int:
    if s.isnumeric(): return s

    a = re.split(r"[\s-]+", s)
    n = 0
    g = 0
    for w in a:
        x = Small.get(w, None)
        if x is not None:
            g += x
        elif w == "hundred" and g != 0:
            g *= 100
        else:
            x = Magnitude.get(w, None)
            if x is not None:
                n += g * x
                g = 0
            else:
                raise NumberException("Unknown number: " + w)
    return n + g



"""
This module handles all fact transformations.

What are fact transformation rules?
In iXBRL filers are allowed to tag textual values like "one million" or "17th of January 2022". In XBRL those facts
would be represented in a normalized manner (1000000, 2022-01-17). To normalize the text values, the iXBRL specification
provides for so-called transformation rules. The transformation rule "zerodash" for example tells us that the tagged
char "-" has a normalized value of 0.

The transformation rules are collected in a so-called transformation rule registry. As of writing SEC Edgar supports
the following registries:

Name: XII Transformation Registry 3
Prefix: ixt
Namespace: http://www.xbrl.org/inlineXBRL/transformation/2015-02-26

Name: XII Transformation Registry 4
Prefix: ixt
Namespace: http://www.xbrl.org/inlineXBRL/transformation/2020-02-12

Name: SEC Specific Transformation Registry
Prefix: ixt-sec
Namespace: http://www.sec.gov/inlineXBRL/transformation/2015-08-31
"""
import re
from math import floor

from xbrl.transformations.text2num import text2num, replace_text_numbers


class TransformationException(Exception):
    """ Generic Exception thrown if an issue was encountered while transforming the fact value """

    def __init__(self, msg):
        Exception.__init__(self, msg)


class RegistryNotSupported(TransformationException):
    """ Exception thrown if the transformation registry is not supported """

    def __init__(self, namespace: str):
        TransformationException.__init__(self, f"The transformation registry {namespace} is currently not supported")


class InvalidTransformation(TransformationException):
    """ Exception thrown if the transformation format is not in the registry """

    def __init__(self, namespace: str, formatCode: str):
        msg = f'The transformation rule "{formatCode}" is not implemented in registry {namespace}'
        TransformationException.__init__(self, msg)


class TransformationNotImplemented(TransformationException):
    """ Exception thrown if the transformation is in the registry but not implemented in py-xbrl """

    def __init__(self, msg):
        TransformationException.__init__(self, msg)


def notImplemented(arg: str) -> str:
    raise TransformationNotImplemented(f"Could not normalize {arg} because the transformation rule is not implemented")


# region normalization maps
monthNorm = {
    'jan': '01',
    'feb': '02',
    'mar': '03',
    'apr': '04',
    'may': '05',
    'jun': '06',
    'jul': '07',
    'aug': '08',
    'sep': '09',
    'sept': '09',
    'oct': '10',
    'nov': '11',
    'dec': '12',
    'january': '01',
    'february': '02',
    'march': '03',
    'april': '04',
    'june': '06',
    'july': '07',
    'august': '08',
    'september': '09',
    'october': '10',
    'november': '11',
    'december': '12'
}

# exchange maps (https://www.sec.gov/info/edgar/specifications/edgarfm-vol2-v59.pdf page 203)
exchangeNorm = {
    'new york stock exchange': 'NYSE',
    'nasdaq global select market': 'NASDAQ',
    'nasdaq stock market': 'NASDAQ',
    'box exchange': 'BOX',
    'nasdaq bx': 'BX',
    'cboe c2 exchange': 'C2',
    'cboe exchange': 'CBOE',
    'chicago stock exchange': 'CHX',
    'cboe byx exchange': 'CboeBYX',
    'cboe bzx exchange': 'CboeBZX',
    'cboe edga exchange': 'CboeEDGA',
    'cboe edgx exchange': 'CboeEDGX',
    'nasdaq gemx': 'GEMX',
    'investors exchange': 'IEX',
    'nasdaq ise': 'ISE',
    'miami international securities exchange': 'MIAX',
    'nasdaq mrx': 'MRX',
    'nyse american': 'NYSEAMER',
    'nyse arca': 'NYSEArca',
    'nyse national': 'NYSENAT',
    'miax pearl': 'PEARL',
    'nasdaq phlx': 'Phlx',
}

stateNormUS = {
    'alabama': 'AL',
    'alaska': 'AK',
    'arizona': 'AZ',
    'arkansas': 'AR',
    'california': 'CA',
    'colorado': 'CO',
    'connecticut': 'CT',
    'delaware': 'DE',
    'florida': 'FL',
    'georgia': 'GA',
    'hawaii': 'HI',
    'idaho': 'ID',
    'illinois': 'IL',
    'indiana': 'IN',
    'iowa': 'IA',
    'kansas': 'KS',
    'kentucky': 'KY',
    'louisiana': 'LA',
    'maine': 'ME',
    'maryland': 'MD',
    'massachusetts': 'MA',
    'michigan': 'MI',
    'minnesota': 'MN',
    'mississippi': 'MS',
    'missouri': 'MO',
    'montana': 'MT',
    'nebraska': 'NE',
    'nevada': 'NV',
    'new hampshire': 'NH',
    'new jersey': 'NJ',
    'new mexico': 'NM',
    'new york': 'NY',
    'north carolina': 'NC',
    'north dakota': 'ND',
    'ohio': 'OH',
    'oklahoma': 'OK',
    'oregon': 'OR',
    'pennsylvania': 'PA',
    'rhode island': 'RI',
    'south carolina': 'SC',
    'south dakota': 'SD',
    'tennessee': 'TN',
    'texas': 'TX',
    'utah': 'UT',
    'vermont': 'VT',
    'virginia': 'VA',
    'washington': 'WA',
    'washington dc': 'DC',
    'west virginia': 'WV',
    'wisconsin': 'WI',
    'wyoming': 'WY',
}


def yearNorm(year: str) -> str:
    if len(year) == 4: return year
    if len(year) == 2: return '19' + year if int(year) > 55 else '20' + year
    raise TransformationException(f'Could not normalize "{year}" to a year')


# endregion normalization maps

# region ixt mappings


def dateDayMonth(arg: str) -> str:
    """ (D)D*(M)M -> --MM-DD """
    seg = re.split(r'[^\d]+', arg)  # split at any char sequence that is not a digit
    return f"--{seg[1].zfill(2)}-{seg[0].zfill(2)}"


def dateDayMonthEN(arg: str) -> str:
    # (D)D*Mon(th) -> --MM-DD
    seg = re.split(r'[^\d\w]+', arg)  # split at any char that is not a digit nor a word
    return f"--{monthNorm[seg[1]]}-{seg[0].zfill(2)}"


def dateDayMonthYear(arg: str) -> str:
    # (D)D*(M)M*(Y)Y(YY) -> YYYY-MM-DD
    seg = re.split(r'[^\d]+', arg)  # split at any char sequence that is not a digit
    return f"{yearNorm(seg[2])}-{seg[1].zfill(2)}-{seg[0].zfill(2)}"


def dateDayMonthYearEN(arg: str) -> str:
    # (D)D*Mon(th)*(Y)Y(YY) -> YYYY-MM-DD
    seg = re.split(r'[^\d\w]+', arg)  # split at any char that is not a digit nor a word
    return f"{yearNorm(seg[2])}-{monthNorm[seg[1]]}-{seg[0].zfill(2)}"


def dateMonthDay(arg: str) -> str:
    # (M)M*(D)D -> --MM-DD
    seg = re.split(r'[^\d]+', arg)  # split at any char sequence that is not a digit
    return f"--{seg[0].zfill(2)}-{seg[1].zfill(2)}"


def dateMonthDayEN(arg: str) -> str:
    # Mon(th)*(D)D -> --MM-DD
    seg = re.split(r'[^\d\w]+', arg)  # split at any char that is not a digit nor a word
    return f"--{monthNorm[seg[0]]}-{seg[1].zfill(2)}"


def dateMonthDayYear(arg: str) -> str:
    # (M)M*(D)D*(Y)Y(YY) -> YYYY-MM-DD
    seg = re.split(r'[^\d]+', arg)  # split at any char sequence that is not a digit
    return f"{yearNorm(seg[2])}-{seg[0].zfill(2)}-{seg[1].zfill(2)}"


def dateMonthDayYearEN(arg: str) -> str:
    # Mon(th)*(D)D*(Y)Y(YY) -> YYYY-MM-DD
    seg = re.split(r'[^\d\w]+', arg)  # split at any char that is not a digit nor a word
    return f"{yearNorm(seg[2])}-{monthNorm[seg[0]]}-{seg[1].zfill(2)}"


def dateMonthYear(arg: str) -> str:
    # (M)M*(Y)Y(YY) -> YYYY-MM
    seg = re.split(r'[^\d]+', arg)  # split at any char sequence that is not a digit
    return f"{yearNorm(seg[1])}-{seg[0].zfill(2)}"


def dateMonthYearEN(arg: str) -> str:
    # Mon(th)*(Y)Y(YY) -> YYYY-MM
    seg = re.split(r'[^\d\w]+', arg)  # split at any char that is not a digit nor a word
    return f"{yearNorm(seg[1])}-{monthNorm[seg[0]]}"


def dateYearMonthDay(arg: str) -> str:
    # (Y)Y(YY)*(M)M*(D)D -> YYYY-MM-DD
    seg = re.split(r'[^\d]+', arg)  # split at any char sequence that is not a digit
    return f"{yearNorm(seg[0])}-{seg[1].zfill(2)}-{seg[2].zfill(2)}"


def dateYearMonth(arg: str) -> str:
    # (Y)Y(YY)*(M)M
    seg = re.split(r'[^\d]+', arg)  # split at any char sequence that is not a digit
    return f"{yearNorm(seg[0])}-{seg[1].zfill(2)}"


def dateYearMonthEN(arg: str) -> str:
    # (Y)Y(YY)*Mon(th) -> YYYY-MM
    seg = re.split(r'[^\d\w]+', arg)  # split at any char that is not a digit nor a word
    return f"{yearNorm(seg[0])}-{monthNorm[seg[1]]}"


def numCommaDecimal(arg: str) -> str:
    # nnn*nnn*nnn,n -> nnnnnnnnn.n
    arg = re.sub(r'[^\d,]+', '', arg)  # remove all chars that are not a digit and not a comma
    return arg.replace(',', '.')


def numDotDecimal(arg: str) -> str:
    # nnn*nnn*nnn.n -> nnnnnnnnn.n
    return re.sub(r'[^\d.]+', '', arg)  # remove all chars that are not a digit and not a dot


# endregion ixt mappings

# region ixt-sec mappings

def durYear(arg: str) -> str:
    # -22.3456 -> -P22Y4M4D (Period: 22 Years, 4 Months, 4 Days)
    pos: bool = float(arg) >= 0
    yearDec: float = abs(float(arg))
    fullYears: int = floor(abs(yearDec))
    fullDays: int = round((yearDec - fullYears) * 365.25)
    fullMonths: int = floor(fullDays / 30.437)
    remainingDays: int = round(fullDays - fullMonths * 30.437)
    return f"{'P' if pos else '-P'}{fullYears}Y{fullMonths}M{remainingDays}D"


def durMonth(arg: str) -> str:
    # 22.3456 -> P22M10D (Period: 22 Months, 10 Days)
    pos: bool = float(arg) >= 0
    monthDec: float = abs(float(arg))
    fullMonths: int = floor(abs(monthDec))
    remainingDays: int = round((monthDec - fullMonths) * 30)
    return f"{'P' if pos else '-P'}{fullMonths}M{remainingDays}D"


def durWordSen(arg: str) -> str:
    value = replace_text_numbers(arg)
    years, months, days = 0, 0, 0
    words: [str] = value.split(' ')
    for x in range(len(words) - 1):
        if not words[x].isnumeric():
            continue
        if 'year' in words[x + 1]:
            years = int(words[x])
        elif 'month' in words[x + 1]:
            months = int(words[x])
        elif 'day' in words[x + 1]:
            days = int(words[x])
    return f'P{years}Y{months}M{days}D'


def numWordSen(arg: str) -> str:
    if arg == 'no' or arg == 'none':
        return '0'
    else:
        arg = arg.replace(' and ', ' ')
        return str(text2num(arg))


def ballotBox(arg: str) -> str:
    if arg == '&#9744;' or arg == '☐':
        return 'false'
    elif arg == '&#9745;' or arg == '☑' or arg == '&#9746;' or arg == '☒':
        return 'true'
    else:
        raise TransformationException("Invalid input for ballotBox transformation rule")


def exchNameEN(arg: str) -> str:
    # "The New York Stock Exchange " -> "NYSE"
    name = arg.lower().strip()
    # remove any commas, points, e.t.c and "inc" or "llc"
    name = re.sub(r'[^\w\s\d]|inc|llc|the', '', name)
    # remove multiple spaces ("  ")
    name = re.sub('r {2,}', ' ', name.strip())
    if name.upper() in exchangeNorm.values():
        return name.upper()
    try:
        return exchangeNorm[name]
    except KeyError:
        raise TransformationException(f'Unknown exchange "{arg}"')


def stateNameEN(arg: str) -> str:
    # "Kentucky" -> "KY"
    name = arg.lower().strip()
    # remove any commas, points e.t.c
    name = re.sub(r'[^\w\s\d]', '', name.strip().lower())
    try:
        return stateNormUS[name]
    except KeyError:
        raise TransformationException(f'Unknown US State "{arg}"')


def entityFilerCategoryEN(arg: str) -> str:
    # large accelerated filer -> Large Accelerated Filer
    arg = arg.lower().strip()
    if arg == 'large accelerated filer':
        return 'Large Accelerated Filer'
    elif arg == 'accelerated filer':
        return 'Accelerated Filer'
    elif arg == 'non-accelerated filer':
        return 'Non-accelerated Filer'
    raise TransformationException(f'Unknown filer category "{arg}"')


# endregion ixt-sec mappings

ixt3 = {
    'booleanfalse': lambda arg: 'false',
    'booleantrue': lambda arg: 'true',
    'calindaymonthyear': notImplemented,
    'datedaymonth': dateDayMonth,
    'datedaymonthdk': notImplemented,
    'datedaymonthen': dateDayMonthEN,
    'datedaymonthyear': dateDayMonthYear,
    'datedaymonthyeardk': notImplemented,
    'datedaymonthyearen': dateDayMonthYearEN,
    'datedaymonthyearin': notImplemented,
    'dateerayearmonthdayjp': notImplemented,
    'dateerayearmonthjp': notImplemented,
    'datemonthday': dateMonthDay,
    'datemonthdayen': dateMonthDayEN,
    'datemonthdayyear': dateMonthDayYear,
    'datemonthdayyearen': dateMonthDayYearEN,
    'datemonthyear': dateMonthYear,
    'datemonthyeardk': notImplemented,
    'datemonthyearen': dateMonthYearEN,
    'datemonthyearin': notImplemented,
    'dateyearmonthday': dateYearMonthDay,
    'dateyearmonthdaycjk': notImplemented,
    'dateyearmonthcjk': notImplemented,
    'dateyearmonthen': dateYearMonthEN,
    'nocontent': lambda arg: '',
    'numcommadecimal': numCommaDecimal,
    'numdotdecimal': numDotDecimal,
    'numdotdecimalin': notImplemented,
    'numunitdecimal': notImplemented,
    'numunitdecimalin': notImplemented,
    'zerodash': lambda arg: '0'
}

ixt4 = {
    'date-day-month': dateDayMonth,
    'date-day-month-year': dateDayMonthYear,
    'date-day-monthname-bg': notImplemented,
    'date-day-monthname-cs': notImplemented,
    'date-day-monthname-da': notImplemented,
    'date-day-monthname-de': notImplemented,
    'date-day-monthname-el': notImplemented,
    'date-day-monthname-en': dateDayMonthEN,
    'date-day-monthname-es': notImplemented,
    'date-day-monthname-et': notImplemented,
    'date-day-monthname-fi': notImplemented,
    'date-day-monthname-fr': notImplemented,
    'date-day-monthname-hr': notImplemented,
    'date-day-monthname-it': notImplemented,
    'date-day-monthname-lv': notImplemented,
    'date-day-monthname-nl': notImplemented,
    'date-day-monthname-no': notImplemented,
    'date-day-monthname-pl': notImplemented,
    'date-day-monthname-pt': notImplemented,
    'date-day-monthname-ro': notImplemented,
    'date-day-monthname-sk': notImplemented,
    'date-day-monthname-sl': notImplemented,
    'date-day-monthname-sv': notImplemented,
    'date-day-monthname-year-bg': notImplemented,
    'date-day-monthname-year-cs': notImplemented,
    'date-day-monthname-year-da': notImplemented,
    'date-day-monthname-year-de': notImplemented,
    'date-day-monthname-year-el': notImplemented,
    'date-day-monthname-year-en': dateDayMonthYearEN,
    'date-day-monthname-year-es': notImplemented,
    'date-day-monthname-year-et': notImplemented,
    'date-day-monthname-year-fi': notImplemented,
    'date-day-monthname-year-fr': notImplemented,
    'date-day-monthname-year-hi': notImplemented,
    'date-day-monthname-year-hr': notImplemented,
    'date-day-monthname-year-it': notImplemented,
    'date-day-monthname-year-nl': notImplemented,
    'date-day-monthname-year-no': notImplemented,
    'date-day-monthname-year-pl': notImplemented,
    'date-day-monthname-year-pt': notImplemented,
    'date-day-monthname-year-ro': notImplemented,
    'date-day-monthname-year-sk': notImplemented,
    'date-day-monthname-year-sl': notImplemented,
    'date-day-monthname-year-sv': notImplemented,
    'date-day-monthroman': notImplemented,
    'date-day-monthroman-year': notImplemented,
    'date-ind-day-monthname-year-hi': notImplemented,
    'date-jpn-era-year-month': notImplemented,
    'date-jpn-era-year-month-day': notImplemented,
    'date-month-day': dateMonthDay,
    'date-month-day-year': dateMonthDayYear,
    'date-month-year': dateMonthYear,
    'date-monthname-day-en': dateMonthDayEN,
    'date-monthname-day-hu': notImplemented,
    'date-monthname-day-lt': notImplemented,
    'date-monthname-day-year-en': dateMonthDayYearEN,
    'date-monthname-year-bg': notImplemented,
    'date-monthname-year-cs': notImplemented,
    'date-monthname-year-da': notImplemented,
    'date-monthname-year-de': notImplemented,
    'date-monthname-year-el': notImplemented,
    'date-monthname-year-en': notImplemented,
    'date-monthname-year-es': notImplemented,
    'date-monthname-year-et': notImplemented,
    'date-monthname-year-fi': notImplemented,
    'date-monthname-year-fr': notImplemented,
    'date-monthname-year-hi': notImplemented,
    'date-monthname-year-hr': notImplemented,
    'date-monthname-year-it': notImplemented,
    'date-monthname-year-nl': notImplemented,
    'date-monthname-year-no': notImplemented,
    'date-monthname-year-pl': notImplemented,
    'date-monthname-year-pt': notImplemented,
    'date-monthname-year-ro': notImplemented,
    'date-monthname-year-sk': notImplemented,
    'date-monthname-year-sl': notImplemented,
    'date-monthname-year-sv': notImplemented,
    'date-monthroman-year': notImplemented,
    'date-year-day-monthname-lv': notImplemented,
    'date-year-month': dateYearMonth,
    'date-year-month-day': dateYearMonthDay,
    'date-year-monthname-day-hu': notImplemented,
    'date-year-monthname-day-lt': notImplemented,
    'date-year-monthname-en': dateYearMonthEN,
    'date-year-monthname-hu': notImplemented,
    'date-year-monthname-lt': notImplemented,
    'date-year-monthname-lv': notImplemented,
    'fixed-empty': lambda arg: '',
    'fixed-false': lambda arg: 'false',
    'fixed-true': lambda arg: 'true',
    'fixed-zero': lambda arg: '0',
    'num-comma-decimal': numCommaDecimal,
    'num-dot-decimal': numDotDecimal,
    'num-unit-decimal': notImplemented,
}

# As defined in https://www.sec.gov/info/edgar/specifications/edgarfm-vol2-v59.pdf
ixt_sec = {
    'duryear': durYear,
    'durmonth': durMonth,
    'durweek': notImplemented,
    'durday': notImplemented,
    'durhour': notImplemented,
    'durwordsen': durWordSen,
    'numwordsen': numWordSen,
    'datequarterend': notImplemented,
    'boolballotbox': ballotBox,
    'exchnameen': exchNameEN,
    'stateprovnameen': stateNameEN,
    'countrynameen': notImplemented,
    'edgarprovcountryen': notImplemented,
    'entityfilercategoryen': entityFilerCategoryEN,
}


def normalize(namespace: str, formatCode: str, value: str) -> str:
    value = value.strip().lower()

    try:
        if namespace == 'http://www.xbrl.org/inlineXBRL/transformation/2015-02-26':
            return ixt3[formatCode](value)
        elif namespace == 'http://www.xbrl.org/inlineXBRL/transformation/2020-02-12':
            return ixt4[formatCode](value)
        elif namespace == 'http://www.sec.gov/inlineXBRL/transformation/2015-08-31':
            return ixt_sec[formatCode](value)
        else:
            raise RegistryNotSupported(namespace)
    except KeyError:
        raise InvalidTransformation(namespace, formatCode)
    except TransformationNotImplemented:
        msg = f'The transformation "{formatCode}" rule of registry {namespace} is currently not implemented in py-xbrl'
        raise TransformationNotImplemented(msg)


if __name__ == '__main__':
    print(normalize(
        'http://www.sec.gov/inlineXBRL/transformation/2015-08-31',
        'numwordsen',
        'one million and two'
    ))



from requests.adapters import HTTPAdapter
from urllib3.util import Retry

import requests
import time
import logging

logger = logging.getLogger(__name__)


class ConnectionManager:
    """
    This class handles http requests. If a requests fails, the method will retry the http requests as many times as specified.
    Between each requests the process will sleep {backoff factor} * (2 ** ({number of total retries} - 1)) seconds between
    failed requests. After the request is successful, the process will sleep delay / 1000 seconds so that it is
    compliant with sec's policy.

    https://www.sec.gov/privacy.htm#security

    """

    def __init__(self, delay: int = 500, retries: int = 5, backoff_factor: float = 0.8, headers: dict = None, logs=True,
                 verify_https: bool = True):
        """

        @param delay: Specifies sleeping time after the request is successful.
        @param retries: How many times a request will be tried before assuming its failure.
        @param backoff_factor: Used to measure time to sleep between failed requests.
            The formula used is {backoff factor} * (2 ** ({number of total retries} - 1))
        @param headers: Headers to use in http request.
        """
        self._delay_ms = delay  # post delay after download
        self._retries = retries
        self._backoff_factor = backoff_factor
        self._headers = headers
        self._session = self._create_session()
        self.logs = logs
        self.verify_https = verify_https
        self.next_try_systime_ms = self._get_systime_ms()  # when can we try next download

        if verify_https is False:
            requests.packages.urllib3.disable_warnings()

    def _get_systime_ms(self):
        return int(time.time() * 1000)

    def download(self, url: str, headers: str):
        # make sure last post-delay elapsed, to rate limit API usage
        time.sleep(max(0, self.next_try_systime_ms - self._get_systime_ms()) / 1000)

        response = self._session.get(url, headers=headers, allow_redirects=True, verify=self.verify_https)
        if self.logs: logger.info(str(response.status_code) + " " + url)

        # no actual delay after last download
        self.next_try_systime_ms = self._get_systime_ms() + self._delay_ms

        return response

    def _create_session(self,
                        status_forcelist: tuple = (500, 502, 503, 504, 403)) -> requests.Session:
        session = requests.Session()
        retry = Retry(
            total=self._retries,
            read=self._retries,
            connect=self._retries,
            backoff_factor=self._backoff_factor,
            status_forcelist=status_forcelist,
        )
        adapter = HTTPAdapter(max_retries=retry)
        session.mount('http://', adapter)
        session.mount('https://', adapter)

        return session
